/*
 *  Filename:  screen.c
 *
 *  Programmer:  Br. David Carlson
 *
 *  Reference:  Joy of C, by Miller & Quilici
 *
 *  Date:  May 14, 1994
 *
 *  Modified:  August 1, 1996 to take into account changes in HP-UX 10.X
 *  which changed the codes received for the arrow key presses to:
 *     up arrow     27 79 65
 *     down arrow   27 79 66
 *     right arrow  27 79 67
 *     left arrow   27 79 68
 *  The old codes were the same except that the middle one was 91.
 *  The new screen.c will accept either the new or the old codes.
 *
 *  This auxiliary file provides screen manipulation functions and
 *  keyboard input functions.
 */

#include <stdio.h>
#include <termio.h>
#include <stdlib.h>          /* for getenv */
#include "screen.h"

struct termio oldt, t;       /* old and new ternminal status */
static char *tn = NULL;      /* terminal name */
static char tc[1024];        /* terminal controls */
static char *ptr = tc;       /* ptr to terminal controls */
static char *clearscreen;    /* clear screen characters */
static char *movecursor;     /* move cursor characters */
static char td[1024];        /* termcap description */


/*  This function must be called once prior to all uses of getch().
 *  It is used to set up for immediate input of one char using getch()
 *  with no echoing of the char to the screen.
 */

getch_init()
   {
   setvbuf(stdin, NULL, _IONBF, NULL);
   if (ioctl(0, TCGETA, &t) == -1)
      return -1;

   oldt = t;
   t.c_lflag &= ~ICANON;
   t.c_lflag &= ~ECHO;
   t.c_cc[4] = 1;
   t.c_cc[5] = 0;

   if (ioctl(0, TCSETAF, &t) != -1)
      return 0;
   else
      return -1;
   }


/*  Call the following function once when finished with getch() in
 *  order to return to normal input operations on stdin.
 */

getch_done()
   {
   setvbuf(stdin, NULL, _IOFBF, BUFSIZ);
   if (ioctl(0, TCSETAF, &oldt) != -1)
      return 0;
   else
      return -1;
   }


/*  The following function clears the screen.  */

display_clear()
   {
   printf(clearscreen);
   }


/*  Call the following function once before using the display_goto()
 *  function to move the cursor to specific locations on the screen.
 */

display_init()
   {
   if (((tn = getenv("TERM")) != NULL) && (tgetent(td, tn) == 1))
      {
      clearscreen = tgetstr("cl", &ptr);
      movecursor = tgetstr("cm", &ptr);
      }
   else
      return -1;
   display_clear();
   return 0;
   }


static int putcontrol(c)
int c;
   {
   return putchar(c);
   }


/*  The following function is used to move the cursor to specific
 *  coordinate (row = r, column = c) on the screen.
 */

display_goto(r, c)
int r, c;
   {
   tputs(tgoto(movecursor, c, r), 1, putcontrol);
   }


/*  The following function is used to get a user's key press.  The
 *  function return value is PRINTABLE_KEY, SPECIAL_KEY, or UNKNOWN_KEY.
 *  The parameter returns the key itself, either the ASCII for the
 *  key, or F1, F2, UP_ARROW, etc for special keys.  (See screen.h
 *  for a complete list of these constants.)
 */

getkey(KeyPtr)
   int *KeyPtr;
   {
   int ch1, ch2, ch3, ch4;

   ch1 = getch();
   if (ch1 == ESC)
      {
      ch2 = getch();
      if (ch2 == 79)
         {
         ch3 = getch();
         if (ch3 == 80)
            *KeyPtr = F1;
         else if (ch3 == 81)
            *KeyPtr = F2;
         else if (ch3 == 82)
            *KeyPtr = F3;
         else if (ch3 == 83)
            *KeyPtr = F4;
         else if (ch3 == 84)
            *KeyPtr = F5;
         else if (ch3 == 85)
            *KeyPtr = F6;
         else if (ch3 == 86)
            *KeyPtr = F7;
         else if (ch3 == 87)
            *KeyPtr = F8;
         else if (ch3 == 88)
            *KeyPtr = F9;
         else if (ch3 == 89)
            *KeyPtr = F10;
         else if (ch3 == 65)
            *KeyPtr = UP_ARROW;
         else if (ch3 == 66)
            *KeyPtr = DOWN_ARROW;
         else if (ch3 == 67)
            *KeyPtr = RIGHT_ARROW;
         else if (ch3 == 68)
            *KeyPtr = LEFT_ARROW;
         else
            return UNKNOWN_KEY;
         }
      else if (ch2 == 91)
         {
         ch3 = getch();
         if (ch3 == 65)
            *KeyPtr = UP_ARROW;
         else if (ch3 == 66)
            *KeyPtr = DOWN_ARROW;
         else if (ch3 == 67)
            *KeyPtr = RIGHT_ARROW;
         else if (ch3 == 68)
            *KeyPtr = LEFT_ARROW;
         else if ((ch3 >= 49) && (ch3 <= 54))
            {
            ch4 = getch();
            if (ch4 != 126)
               return UNKNOWN_KEY;
            if (ch3 == 49)
               *KeyPtr = INSERT;
            else if (ch3 == 50)
               *KeyPtr = HOME;
            else if (ch3 == 51)
               *KeyPtr = PAGE_UP;
            else if (ch3 == 52)
               *KeyPtr = DELETE;
            else if (ch3 == 53)
               *KeyPtr = END;
            else if (ch3 == 54)
               *KeyPtr = PAGE_DOWN;
            }
         else
            return UNKNOWN_KEY;
         }
      else
         return UNKNOWN_KEY;
      }
   else if (((ch1 >= SPACE) && (ch1 <= BACKSPACE))
      || (ch1 == TAB) || (ch1 == ENTER) || (ch1 == CTRL_R) || (ch1 == CTRL_Y)
      || (ch1 == CTRL_T) || (ch1 == CTRL_B) || (ch1 == CTRL_E))
      {
      *KeyPtr = ch1;
      }
   else
      return UNKNOWN_KEY;

   if ((*KeyPtr >= BACKSPACE) || (*KeyPtr == ENTER) || (*KeyPtr == TAB)
      || (*KeyPtr == CTRL_R) || (*KeyPtr == CTRL_Y) || (*KeyPtr == CTRL_B)
      || (*KeyPtr == CTRL_E) || (*KeyPtr == CTRL_T))
      return SPECIAL_KEY;
   else
      return PRINTABLE_KEY;
   }

