/*
 *  Filename:  util.c
 *
 *  Programmer:  Br. David Carlson
 *
 *  Date:  June 22, 1994
 *
 *  This is one of the source files for the ted editor project.  This file
 *  contains various utility functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "screen.h"
#include "ted.h"


/* Global variables for quick access to commonly used items: */

extern LINE_TYPE *buffer;
extern MARK_RECORD mark;
extern char source_filename[80];
extern INT8 mode, proceed, insert, BlockMarked;
extern INT16 line_count, avail, first, last, current, file_row,
   screen_row, screen_col, file_col;
extern INT32 buffer_size;


/*  The following function beeps until it gets an ENTER from the user. */

void get_enter(void)
   {
   INT8 code;
   int key;

   code = (INT8) getkey(&key);
   while ((code != SPECIAL_KEY) || (key != ENTER))
      {
      beep();
      code = (INT8) getkey(&key);
      }
   }


/*  The following function beeps until it gets ch1 or ch2 as the keypress.
 *  Note that ch1 and ch2 must be in lower case.
 */

char get_choice(char ch1, char ch2)
   {
   INT8 code;
   int key;

   code = (INT8) getkey(&key);
   key = tolower(key);

   while (! ((code == PRINTABLE_KEY) && ((key == ch1) || (key == ch2))))
      {
      beep();
      code = (INT8) getkey(&key);
      key = tolower(key);
      }

   return (char) key;
   }


/*  The following function beeps until it gets a key press of F3 or a valid
 *  PgUp of PgDn key press, depending on what is allowed according to the
 *  current value pointed to by the parameter.  The key press is used to 
 *  determine the new screen number to be returned via the parameter.  A
 *  value of 0 is returned if F3 is pressed.
 */

void get_screen_num(INT8 *screen_num_ptr)
   {
   INT8 code;
   int key;

   code = (INT8) getkey(&key);

   while (! ((code == SPECIAL_KEY) && ((key == F3) ||
      ((key == PAGE_UP) && (*screen_num_ptr > 1)) ||
      ((key == PAGE_DOWN) && (*screen_num_ptr < 3)))))
      {
      beep();
      code = (INT8) getkey(&key);
      }

   if (key == PAGE_UP)
      (*screen_num_ptr)--;
   else if (key == PAGE_DOWN)
      (*screen_num_ptr)++;
   else
      *screen_num_ptr = 0;
   }


/*  The following function reads a string from the keyboard.  It allows
 *  the use of the usual editing keys.  The string can contain up to max
 *  characters.  Input begins on the screen at row, start_col.
 */

INT16 get_string(char s[], INT16 max, INT16 row, INT16 start_col)
   {
   register int k;
   INT8 code, local_insert;
   int key;
   INT16 len, col;

   display_goto(row, start_col);
   len = 0;
   col = 0;
   local_insert = TRUE;

   while (TRUE)
      {
      code = (INT8) getkey(&key);
      if (code == PRINTABLE_KEY)
         {
         if ((len < max) && (local_insert))
            {
            for (k = len; k > col; k--)
               s[k] = s[k - 1];
            s[col] = (char) key;
            len++;
            for (k = col; k < len; k++)
               putchar(s[k]);
            col++;
            display_goto(row, start_col + col);
            }
         else if ((col < max) && (! local_insert))
            {
            s[col++] = (char) key;
            putchar(key);
            if (col > len)
               len = col;
            }
         else
            beep();
         }
      else if (code == SPECIAL_KEY)
         {
         if (key == ENTER)
            break;
         else if (key == BACKSPACE)
            {
            if (col > 0)
               {
               col--;
               display_goto(row, start_col + col);
               if (col < len)
                  {
                  len--;
                  for (k = col; k < len; k++)
                     {
                     s[k] = s[k + 1];
                     putchar(s[k]);
                     }
                  putchar(' ');
                  display_goto(row, start_col + col);
                  }
               }
            else
               beep();
            }
         else if (key == DELETE)
            {
            if (col < len)
               {
               len--;
               for (k = col; k < len; k++)
                  {
                  s[k] = s[k + 1];
                  putchar(s[k]);
                  }
               s[len] = ' ';
               putchar(' ');
               display_goto(row, start_col + col);
               }
            }
         else if (key == LEFT_ARROW)
            {
            if (col > 0)
               {
               col--;
               display_goto(row, start_col + col);
               }
            else
               beep();
            }
         else if (key == RIGHT_ARROW)
            {
            if (col < max)
               {
               col++;
               display_goto(row, start_col + col);
               }
            else
               beep();
            }
         else if (key == INSERT)
            local_insert = 1 - local_insert;
         else if (key == HOME)
            {
            col = 0;
            display_goto(row, start_col);
            }
         else if (key == END)
            {
            col = len;
            display_goto(row, start_col + col);
            }
         else
            beep();
         }
      else   /* unknown key */
         beep();
      }

   return len;
   }



/*  The following function figures out the buffer index for the line
 *  at position row on the screen.  It uses the global current (the current
 *  buffer index) and screen_row (the current row on the screen).  NIL
 *  will be returned if there is no line at position row on the screen.
 */

INT16 get_index(INT16 row)
   {
   register int k, index;
   INT16 difference;

   index = current;
   difference = row - screen_row;

   if (difference < 0)
      {
      for (k = 0; k > difference; k--)
         {
         index = buffer[index].left;
         if (index == NIL)
            return NIL;
         }
      }
   else
      {
      for (k = 0; k < difference; k++)
         {
         index = buffer[index].right;
         if (index == NIL)
            return NIL;
         }
      }
   return index;
   }


/*  The following function displays the buffer line with the given index
 *  starting at location row, col on the screen, but only the number of
 *  characters given by num_cols.
 */

void print_line(INT16 index, INT16 row, INT16 col, INT16 num_cols)
   {
   register int k;
   LINE_TYPE *buf_ptr;

   display_goto(row, col);

   if (index == NIL)
      {
      display_color(NORMAL);
      for (k = 0; k < num_cols; k++)
         putchar(' ');
      }
   else
      {
      buf_ptr = &buffer[index];
      if (buf_ptr->marked)
         display_color(CYAN);
      else
         display_color(NORMAL);

      for (k = col; k < col + num_cols; k++)
         putchar(buf_ptr->info[k]);

      if (buf_ptr->marked)
         display_color(NORMAL);
      }
   }


/*  The following function removes the window starting at screen location
 *  row, col and extending for num_rows and num_cols.  It replaces the
 *  window area with the normal text (or blanks) that had been there.
 */

void remove_window(INT16 row, INT16 num_rows, INT16 col, INT16 num_cols)
   {
   register int k;
   INT16 index;

   index = get_index(row);
   display_color(NORMAL);

   for (k = row; k < row + num_rows; k++)
      {
      print_line(index, k, col, num_cols);
      if (index != NIL)
         index = buffer[index].right;
      }

   display_goto(screen_row, screen_col);
   }


/*  The following function displays the screen starting at row and
 *  extending until end_row is reached.  The starting index for the first
 *  row must be passed in.
 */

void display_text(INT16 row, INT16 end_row, INT16 index)
   {
   register int k;

   for (k = row; k <= end_row; k++)
      {
      print_line(index, k, 0, 80);
      if (index != NIL)
         index = buffer[index].right;
      }

   display_goto(screen_row, screen_col);
   }


/*  The following procedure displays all 22 lines of the screen so that the
 *  current line appears in TargetScreenRow, if possible.
 */

void display_text_adjusted(INT16 TargetScreenRow)
   {
   register int count;
   INT16 index, DistToBottom, DistToTop, next_index;

   if ((index = current) != NIL)
      {
      DistToBottom = 23 - TargetScreenRow;
      for (count = 0; count < DistToBottom; count++)
         {
         index = buffer[index].right;
         if (index == NIL)
            break;
         }

      TargetScreenRow += DistToBottom - count;
      DistToTop = TargetScreenRow - 2;

      index = current;
      for (count = 0; count < DistToTop; count++)
         if ((next_index = buffer[index].left) == NIL)
            break;
         else
            index = next_index;
      
      screen_row = TargetScreenRow - DistToTop + count;
      }   

   display_text(2, 23, index);
   show_row_col();
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function allocates one line from the buffer's available
 *  list, returning the index number for the line.  If no line is free,
 *  NIL is returned.
 */

INT16 get_buffer_line(void)
   {
   register int k;
   INT16 temp;
   LINE_TYPE *buf_ptr;

   temp = avail;
   if (temp == NIL)
      return NIL;
   else
      {
      buf_ptr = &buffer[temp];
      avail = buf_ptr->right;
      buf_ptr->length = 0;
      buf_ptr->marked = FALSE;
      for (k = 0; k < 80; k++)
         buf_ptr->info[k] = ' ';
      return temp;
      }
   }


/*  The following function returns the buffer line given by index
 *  to the list of available lines.
 */

void return_buffer_line(INT16 index)
   {
   buffer[index].right = avail;
   avail = index;
   }


/*  The following function displays an error message saying that the buffer
 *  is out of room.
 */

void allocation_error(void)
   {
   beep();
   display_color(NORMAL);
   display_goto(8, 13);
   display_color(RED);
   printf(" ERROR:  The editor's buffer has no more room.        ");
   display_goto(9, 13);
   printf(" You may first have to repeatedly remove some lines.  ");
   display_goto(10, 13);
   printf(" Press ENTER                                          ");
   display_goto(10, 26);
   get_enter();
   remove_window(8, 3, 13, 54);
   }


/*  The following function reports various errors that retrieval might
 *  encounter.
 */

void report_error(INT8 problem)
   {
   beep();
   display_color(NORMAL);
   display_goto(11, 31);
   display_color(RED);
   if (problem == OPENERROR)
      printf(" Cannot open file ");
   else if (problem == READERROR)
      printf(" Read error       ");
   else if (problem == LENGTHERROR)
      printf(" Line too long    ");
   else if (problem == ALLOCERROR)
      printf(" Buffer filled    ");
   display_goto(12, 31);
   printf(" Retrieve aborted ");
   display_goto(13, 31);
   printf(" Press ENTER      ");
   display_goto(13, 44);
   get_enter();
   remove_window(11, 3, 31, 18);
   }


/*  The following function reads the lines from the source file into
 *  the buffer and counts the number of lines.  It also quickly displays
 *  the first 22 lines on the screen.  The buffer is initialized as a
 *  doubly linked list of lines.  The available list is adjusted.
 */

void read_file(FILE *source)
   {
   register int ch, col;
   INT16 row, index, prev, old_prev, new_col;
   INT8 last_was_newline, first_ch;
   LINE_TYPE *buf_ptr;

   row = 0;
   col = 0;
   if ((index = get_buffer_line()) == NIL)
      ERROR("Unable to allocate memory space")

   current = index;
   first = index;
   prev = index;
   old_prev = NIL;
   buf_ptr = &buffer[index];
   buf_ptr->left = NIL;
   first_ch = TRUE;
   last_was_newline = FALSE;

   while (!feof(source))
      {
      ch = fgetc(source);
      if (ferror(source))
         ERROR("Error in reading original file")

      if (ch == '\n')
         {
         buf_ptr->length = col;
         /* must pad with blanks to end of standard line length */
         for ( ; col < 80; col++)
            buf_ptr->info[col] = ' ';
         row++;
         col = 0;
         if ((index = get_buffer_line()) == NIL)
            ERROR("Insufficient buffer space")

         buf_ptr = &buffer[index];
         if (row < 22)
            display_goto(2 + row, 0);

         buffer[prev].right = index;
         buf_ptr->left = prev;
         buf_ptr->right = NIL;
         old_prev = prev;
         prev = index;
         last_was_newline = TRUE;
         }
      else if (ch == EOF)
         {
         if (last_was_newline)
            {
            return_buffer_line(index);
            index = old_prev;
            buffer[index].right = NIL;
            row--;
            }
         else if (first_ch)   /* we have an empty file */
            {
            buf_ptr->right = NIL;
            row = 0;
            }
         else
            {   /* incomplete last line */
            buf_ptr->length = col;
            for ( ; col < 80; col++)
               buf_ptr->info[col] = ' ';
            }
         break;
         }
      else if (ch == '\t')
         {
         new_col = 3 * (col / 3) + 3;
         if (new_col > 80)
            ERROR("Line longer than 80 characters encountered")
         for ( ; col < new_col; col++)
            {
            buf_ptr->info[col] = ' ';
            if (row < 22)
               putchar(' ');
            }
         last_was_newline = FALSE;
         }
      else   /* ch is an ordinary character */
         {
         if (col > 79)
            ERROR("Line longer than 80 characters encountered")
         buf_ptr->info[col++] = ch;
         last_was_newline = FALSE;
         if (row < 22)
            putchar(ch);
         }

      first_ch = FALSE;
      }

   last = index;
   line_count = row + 1;
   }


/*  The following function creates a backup name string identical to
 *  the original name except with .bak appended.
 */

void get_backup_name(char *original_name, char *backup_name)
   {
   register int k, ch;

   k = 0;
   ch = original_name[k];
   while (ch != NULL)
      {
      backup_name[k++] = ch;
      ch = original_name[k];
      }
   backup_name[k++] = '.';
   backup_name[k++] = 'b';
   backup_name[k++] = 'a';
   backup_name[k++] = 'k';
   backup_name[k] = NULL;
   }


/*  The following function copies lines from the buffer (from start_line to
 *  end_line) to the indicated destination file.  Returns TRUE of FALSE to
 *  indicates if it succeeds.
 */

int buffer_to_file(FILE *destination, INT16 start_line, INT16 end_line)
   {
   register int col, count;
   INT16 line_ptr;
   LINE_TYPE *buf_ptr;

   line_ptr = start_line;
   if (line_ptr == NIL)
      return TRUE;
   if ((start_line == end_line) && (buffer[start_line].length == 0))
      return TRUE;

   count = line_count;
   while (TRUE)
      {
      buf_ptr = &buffer[line_ptr];
      for (col = 0; col < buf_ptr->length; col++)
         {
         if (fputc(buf_ptr->info[col], destination) == EOF)
            return FALSE;
         }

      if (fputc('\n', destination) == EOF)
         return FALSE;
      count--;
      if ((line_ptr == end_line) || (count == 0))   /* redundant, but */
         break;             /* safer -- be sure no infinite loops */
      line_ptr = buf_ptr->right;
      }

   return TRUE;
   }


/*  The following routine initializes the buffer to empty and sets up
 *  the list of available lines to contain all lines of the buffer.
 *  The available list is a singly-linked list (using the right field).
 */

void buf_initialize(void)
   {
   register int k;
   INT32 last_line;

   first = NIL;
   last = NIL;
   avail = 0;
   last_line = buffer_size - 1;

   for (k = 0; k < last_line; k++)
      buffer[k].right = k + 1;

   buffer[last_line].right = NIL;
   }


/*  The following function prints a line of 80 blanks.  */

void print_blank_line(void)
   {
   register int k;

   for (k = 0; k < 80; k++)
      putchar(' ');
   }


/*  The following function displays the current file column setting. */

void show_file_col(void)
   {
   display_color(NORMAL);
   display_color(WHITE);
   display_goto(1, 78);
   printf("%2d", file_col);
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function displays the current file row setting. */

void show_file_row(void)
   {
   display_color(NORMAL);
   display_color(WHITE);
   display_goto(1, 65);
   printf("%4d", file_row);
   display_goto(screen_row, screen_col);
   display_color(NORMAL);
   }


/*  The following function shows the current file_row and file_col settings.
 *  It should be followed by moving the cursor and setting screen color as
 *  needed.
 */

void show_row_col(void)
   {
   display_color(NORMAL);
   display_color(WHITE);
   display_goto(1, 65);
   printf("%4d", file_row);
   display_goto(1, 78);
   printf("%2d", file_col);
   }


/* The following function displays the first 2 lines of the screen.
 * These are the information lines.
 */

void display_info_lines(void)
   {
   display_color(BLACK);
   display_goto(0, 0);
   printf("F1 Help  F2 Srch  F3 Exit  F4 Mark  F5 Cut  F6 Copy  ");
   printf("F7 Paste  F8 Retrv  F9 Save");
   display_color(NORMAL);

   display_color(WHITE);
   display_goto(1, 0);
   print_blank_line();
   display_goto(1, 0);
   if (insert)
      printf("Insert");
   else
      printf("Typeover");

   display_goto(1, 14);
   printf("%s", source_filename);
   display_goto(1, 60);
   printf("Line %4d", file_row);
   display_goto(1, 74);
   printf("Col %2d", file_col);
   display_color(NORMAL);
   display_goto(2, 0);
   }


/*  The following function unmarks the currently marked block. */

void unmark(void)
   {
   register int k;
   INT16 index;

   BlockMarked = FALSE;
   mark.num_marks = 0;
   index = mark.first_buf_line;

   for (k = mark.first_file_row; k <= mark.second_file_row; k++)
      {
      buffer[index].marked = FALSE;
      index = buffer[index].right;
      }
   }

